use crate::{ser::Serializable, varint::VarInt};

#[derive(Clone, Debug)]
pub struct Commands(Vec<Node>);

impl Commands {
	pub fn new() -> Self {
		let root = Node { children: Vec::new(), data: NodeData::Root };
		let basic = Node {
			children: Vec::new(),
			data: NodeData::Argument {
				redirect: None,
				name: "args".to_owned(),
				parser: Parser::String(StringType::GreedyPhrase),
				suggestion: None,
				executable: true,
			}
		};
		Self(vec![root, basic])
	}

	pub fn add_node(&mut self, data: NodeData) -> Result<i32, &'static str> {
		if let NodeData::Root = data {
			return Err("node data cannot be a root")
		}
		if let NodeData::Argument { redirect: Some(redir), .. } = data {
			if redir < 0 || (redir as usize) >= self.0.len() {
				return Err("invalid redirect")
			}
		}
		if let NodeData::Literal { redirect: Some(redir), .. } = data {
			if redir < 0 || (redir as usize) >= self.0.len() {
				return Err("invalid redirect")
			}
		}
		let node = Node {
			children: Vec::new(),
			data,
		};
		self.0.push(node);
		Ok(self.0.len() as i32 - 1)
	}

	pub fn add_child(&mut self, parent: i32, child: i32) -> Result<(), &'static str> {
		if parent < 0 || (parent as usize) >= self.0.len() {
			return Err("invalid parent node")
		}
		if child < 0 || (child as usize) >= self.0.len() {
			return Err("invalid child node")
		}
		self.0[parent as usize].children.push(child);
		Ok(())
	}

	pub fn add_simple_command(&mut self, name: String) -> Result<(), &'static str> {
		let new_id = self.add_node(NodeData::Literal {
			redirect: None,
			name,
			executable: true
		})?;
		self.add_child(Commands::root(), new_id)?;
		self.add_child(new_id, Commands::basic_arg())?;
		Ok(())
	}

	pub fn len(&self) -> i32 {
		self.0.len() as i32
	}

	pub fn serialize(&self, w: &mut impl std::io::Write, new: bool) -> anyhow::Result<()> {
		for node in &self.0 {
			node.serialize(w, new)?;
		}
		Ok(())
	}

	pub const fn root() -> i32 { 0 }

	pub const fn basic_arg() -> i32 { 1 }
}

#[derive(Clone, Debug)]
struct Node {
	children: Vec<i32>,
	data: NodeData,
}

impl Node {
    pub fn serialize(&self, w: &mut impl std::io::Write, new: bool) -> anyhow::Result<()> {
        match &self.data {
            NodeData::Root => {
				// flags (root node)
				0u8.serialize(w)?;
				// children
				VarInt(self.children.len() as i32).serialize(w)?;
				for child in &self.children {
					VarInt(*child).serialize(w)?;
				}
			},
            NodeData::Literal { redirect, name, executable } => {
				// flags:
				let flags: u8 = 0x01 // literal node
					| (*executable as u8 * 0x04) // is executable
					| (redirect.is_some() as u8 * 0x08); // has redirect
				flags.serialize(w)?;
				// children
				VarInt(self.children.len() as i32).serialize(w)?;
				for child in &self.children {
					VarInt(*child).serialize(w)?;
				}
				// redirect
				if let Some(redir) = redirect {
					VarInt(*redir).serialize(w)?;
				}
				// name
				name.serialize(w)?;
			},
            NodeData::Argument { redirect, name, parser, suggestion, executable } => {
				// flags:
				let flags: u8 = 0x02 // argument node
					| (*executable as u8 * 0x04) // is executable
					| (redirect.is_some() as u8 * 0x08) // has redirect
					| (suggestion.is_some() as u8 * 0x10); // has suggestion
				flags.serialize(w)?;
				// children
				VarInt(self.children.len() as i32).serialize(w)?;
				for child in &self.children {
					VarInt(*child).serialize(w)?;
				}
				// redirect
				if let Some(redir) = redirect {
					VarInt(*redir).serialize(w)?;
				}
				// name
				name.serialize(w)?;
				// parser
				parser.serialize(w, new)?;
				// suggestion
				if let Some(sug) = suggestion {
					sug.serialize(w)?;
				}
			},
        }
		Ok(())
    }
}

#[derive(Clone, Debug)]
pub enum NodeData {
	Root,
	Literal {
		redirect: Option<i32>,
		name: String,
		executable: bool,
	},
	Argument {
		redirect: Option<i32>,
		name: String,
		parser: Parser,
		suggestion: Option<Suggestion>,
		executable: bool,
	},
}

#[derive(Clone, Copy, Debug)]
pub enum Suggestion {
	AskServer
}

impl Serializable for Suggestion {
    fn serialize(&self, w: &mut impl std::io::Write) -> anyhow::Result<()> {
        match self {
            Suggestion::AskServer => "minecraft:ask_server".serialize(w)
        }
    }
}

#[derive(Clone, Copy, Debug)]
pub enum Parser {
	Bool,
	Float { min: Option<f32>, max: Option<f32> },
	Double { min: Option<f64>, max: Option<f64> },
	Integer { min: Option<i32>, max: Option<i32> },
	Long { min: Option<i64>, max: Option<i64> },
	String(StringType),
}

impl Parser {
    pub fn serialize(&self, w: &mut impl std::io::Write, new: bool) -> anyhow::Result<()> {
		match (self, new) {
			(Parser::Bool, true)  => VarInt(0).serialize(w)?,
			(Parser::Bool, false) => "brigadier:bool".serialize(w)?,
			(Parser::Float { .. }, true)  => VarInt(1).serialize(w)?,
			(Parser::Float { .. }, false) => "brigadier:float".serialize(w)?,
			(Parser::Double { .. }, true)  => VarInt(2).serialize(w)?,
			(Parser::Double { .. }, false) => "brigadier:double".serialize(w)?,
			(Parser::Integer { .. }, true)  => VarInt(3).serialize(w)?,
			(Parser::Integer { .. }, false) => "brigadier:integer".serialize(w)?,
			(Parser::Long { .. }, true)  => VarInt(4).serialize(w)?,
			(Parser::Long { .. }, false) => "brigadier:long".serialize(w)?,
			(Parser::String(_), true)  => VarInt(5).serialize(w)?,
			(Parser::String(_), false) => "brigadier:string".serialize(w)?,
		};
        match self {
            Parser::Bool => (),
            Parser::Float { min, max } => {
				let flags: u8 = (min.is_some() as u8) | (max.is_some() as u8 * 0x02);
				flags.serialize(w)?;
				if let Some(min) = min {
					min.serialize(w)?;
				}
				if let Some(max) = max {
					max.serialize(w)?;
				}
			}
            Parser::Double { min, max } => {
				let flags: u8 = (min.is_some() as u8) | (max.is_some() as u8 * 0x02);
				flags.serialize(w)?;
				if let Some(min) = min {
					min.serialize(w)?;
				}
				if let Some(max) = max {
					max.serialize(w)?;
				}
			}
            Parser::Integer { min, max } => {
				let flags: u8 = (min.is_some() as u8) | (max.is_some() as u8 * 0x02);
				flags.serialize(w)?;
				if let Some(min) = min {
					min.serialize(w)?;
				}
				if let Some(max) = max {
					max.serialize(w)?;
				}
			}
            Parser::Long { min, max } => {
				let flags: u8 = (min.is_some() as u8) | (max.is_some() as u8 * 0x02);
				flags.serialize(w)?;
				if let Some(min) = min {
					min.serialize(w)?;
				}
				if let Some(max) = max {
					max.serialize(w)?;
				}
			}
            Parser::String(ty) => {
				match ty {
                    StringType::SingleWord => VarInt(0),
                    StringType::QuotablePhrase => VarInt(1),
                    StringType::GreedyPhrase => VarInt(2),
                }.serialize(w)?;
			}
        }
		Ok(())
    }
}

#[derive(Clone, Copy, Debug)]
pub enum StringType {
	SingleWord,
	QuotablePhrase,
	GreedyPhrase,
}
