use std::sync::Arc;

use log::warn;
use mlua::{Lua, Function, IntoLua, FromLua, Value, Table, AnyUserData};
use tokio::sync::{mpsc::{Sender, Receiver}, RwLock};
use uuid::Uuid;

use crate::command::Commands;

use self::{event::{PluginEvent, ServerEvent}, server::{Server, WrappedServer}, plugin::load_plugins};

pub mod event;
mod plugin;
mod server;

#[derive(Clone, Copy, Eq, PartialEq)]
struct UuidUD(Uuid);

impl<'lua> FromLua<'lua> for UuidUD {
	fn from_lua(value: mlua::Value<'lua>, _lua: &'lua Lua) -> mlua::Result<Self> {
		match value {
			Value::UserData(ud) => Ok(*ud.borrow::<Self>()?),
			v => Err(mlua::Error::FromLuaConversionError { from: v.type_name(), to: "Uuid", message: Some("Expected Uuid".to_owned()) })
		}
	}
}

impl mlua::UserData for UuidUD {
	fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
		methods.add_meta_method("__tostring",
			|lua, this, ()| this.0.to_string().into_lua(lua)
		);
		methods.add_meta_method("__eq",
			|lua, this, UuidUD(uuid)| (this.0 == uuid).into_lua(lua)
		);
	}
}

fn broadcast_event<'lua, F>(lua: &'lua Lua, prio: &[String], fname: &str, callback: F) -> mlua::Result<()>
where F: Fn(&str, Table<'lua>, AnyUserData, Function) -> mlua::Result<Option<bool>> {
	let srv: AnyUserData = lua.globals().get("srv")?;
	let plugins: Table = srv.nth_user_value(1)?;
	for id in prio {
		let table: Table = plugins.get(id.clone())?;
		if let Ok(f) = table.get::<_, Function>(fname) {
			srv.set_nth_user_value(2, id.clone())?;
			let result = callback(id, table.clone(), srv.clone(), f);
			match result {
				Ok(Some(true)) => return Ok(()),
				Ok(_) => (),
				Err(e) => warn!("error in {fname} in plugin {}: {e}", id),
			}
		}
	}
	Ok(())
}

fn run_event<'lua, F>(lua: &'lua Lua, fname: &str, id: &str, callback: F) -> mlua::Result<Option<mlua::Error>>
where F: FnOnce(Table<'lua>, AnyUserData, Function) -> mlua::Result<()> {
	let srv: AnyUserData = lua.globals().get("srv")?;
	let plugins: Table = srv.nth_user_value(1)?;
	let table: Table = plugins.get(id)?;
	if let Ok(f) = table.get::<_, Function>(fname) {
		srv.set_nth_user_value(2, id.clone())?;
		if let Err(e) = callback(table.clone(), srv.clone(), f) {
			warn!("error in {fname} in plugin {}: {e}", id);
			return Ok(Some(e));
		}
	}
	Ok(None)
}

pub async fn run_plugins(tx: Sender<PluginEvent>, mut rx: Receiver<ServerEvent>, commands: Arc<RwLock<Commands>>) -> mlua::Result<()> {
	let lua = unsafe {
		mlua::Lua::unsafe_new_with(
			mlua::StdLib::ALL_SAFE | mlua::StdLib::DEBUG,
			mlua::LuaOptions::default()
		)
	};

	let server = WrappedServer(Arc::new(Server {
		tx,
		commands,
		command_ids: Default::default(),
		names: Default::default(),
		info: Default::default(),
	}));

	let prio;
	{
		let server_any: AnyUserData = FromLua::from_lua(IntoLua::into_lua(server.clone(), &lua)?, &lua)?;
		let (plugins, priorities) = load_plugins(&lua)?;
		prio = priorities;

		server_any.set_nth_user_value(1, plugins)?;
		server_any.set_nth_user_value(2, Value::Nil)?;
		lua.globals().set("srv", server_any.clone())?;
	}

	broadcast_event(&lua, &prio, "init", |_, pl, _, f| {
		f.call(pl)
	})?;

	server.tx.send(PluginEvent::UpdateCommands).await.unwrap();

	while let Some(ev) = rx.recv().await {
		match ev {
			ServerEvent::Join(player, conn_info) => {
				server.names.write().await.insert(player.uuid, player.name.clone());
				server.info.write().await.insert(player.uuid, conn_info);
				broadcast_event(&lua, &prio, "on_join", |_, pl, _, f| {
					f.call((
						pl,
						UuidUD(player.uuid),
						player.name.clone()
					))
				})?;
			},
			ServerEvent::Leave(player) => {
				broadcast_event(&lua, &prio, "on_leave", |_, pl, _, f| {
					f.call((
						pl,
						UuidUD(player.uuid),
						player.name.clone()
					))
				})?;
				server.names.write().await.remove(&player.uuid);
				server.info.write().await.remove(&player.uuid);
			},
			ServerEvent::Message(player, msg) => {
				broadcast_event(&lua, &prio, "on_message", |_, pl, _, f| {
					f.call((
						pl,
						msg.clone(),
						UuidUD(player.uuid),
						player.name.clone()
					))
				})?;
			},
			ServerEvent::Command(sender, msg) => {
				let msg2 = msg.clone();
				let mut parts = msg2.split_whitespace();
				let cmd = parts.next().unwrap_or_default();

				if let Some(id) = server.command_ids.read().await.get(cmd) {
					let cmd = cmd.strip_prefix(&(id.clone() + ":")).unwrap_or(cmd);

					let args = parts.map(|s| s.to_owned()).collect::<Vec<_>>();
					let result = run_event(&lua, "on_command", id, |pl, _, f| {
						f.call::<_, ()>((
							pl,
							cmd.to_owned(),
							args,
							msg,
							UuidUD(sender.uuid),
							sender.name,
						))
					})?;
					if result.is_some() {
						let msg = serde_json::json!{{
							"text": "An internal error occured while trying to execute that command",
							"color": "red",
						}};
						server.tx.send(PluginEvent::System(msg, Some(sender.uuid)))
							.await.unwrap();
					}

				} else {
					// TODO couldn't do vanilla localization keys, so
					// make this easier for plugins to modify/localize
					let msg = serde_json::json!{[
						{
							"text": "Unknown command: ",
							"color": "red",
						},
						{
							"text": cmd,
							"color": "gold",
						}
					]};
					server.tx.send(PluginEvent::System(msg, Some(sender.uuid)))
						.await.unwrap();
				}
			}
		}
	}
	Ok(())
}
