use std::{fs::{DirEntry, self}, io::ErrorKind};

use anyhow::anyhow;
use log::{debug, warn, info};
use mlua::{Lua, Table};

use crate::QC_VERSION;

struct PluginData<'lua> {
	id: String,
	priority: i32,
	table: Table<'lua>,
}

fn load_plugin<'lua>(entry: &DirEntry, lua: &'lua Lua) -> anyhow::Result<PluginData<'lua>> {
	let ty = entry.file_type()?;
	let mut path = entry.path();
	if ty.is_dir() {
		path.push("main.lua");
	}
	let chunk = lua.load(path);
	debug!("evaluating plugin");
	let table: Table = chunk.eval()?;

	let id = table.get("id")
		.map_err(|e| anyhow!("could not get plugin id: {e}"))?;

	let priority = table.get::<_, Option<i32>>("priority")
		.map_err(|_| anyhow!("invalid priority value"))?
		.unwrap_or(50);

	if !(0..=100).contains(&priority) {
		return Err(anyhow!("priority {priority} out of range 0..=100"))
	}

	if let Ok(qc_ver) = table.get::<_, String>("server_version") {
		debug!("checking plugin version");
		let req = semver::VersionReq::parse(&qc_ver)?;
		if !req.matches(&semver::Version::parse(QC_VERSION).unwrap()) {
			return Err(anyhow!("incompatible with quectocraft version {QC_VERSION}: expected {req}"));
		}
	} else {
		warn!("plugin '{id}' is missing a version reqirement");
	}

	Ok(PluginData { id, priority, table })
}

pub fn load_plugins(lua: &Lua) -> mlua::Result<(Table, Vec<String>)> {
	let plugins = lua.create_table()?;
	debug!("loading plugins");
	let entries = match fs::read_dir("./plugins") {
		Ok(n) => n,
		Err(e) if e.kind() == ErrorKind::NotFound => {
			warn!("plugins directory does not exist, creating");
			if let Err(e) = fs::create_dir("./plugins") {
				warn!("failed to create plugins directory: {e}");
			}
			return Ok((plugins, vec![]))
		},
		Err(e) => {
			warn!("failed to load plugins: {e}");
			return Ok((plugins, vec![]))
		}
	};
	let mut priorities = Vec::new();
	for entry in entries {
		let entry = match entry {
			Ok(e) => e,
			Err(e) => {
				warn!("error reading entry: {e}");
				continue
			}
		};
		let path = entry.path();
		let spath = path.to_str().unwrap_or("<non-unicode path>");
		debug!("loading plugin at {spath}");
		match load_plugin(&entry, lua) {
			Ok(plugin) => {
				if let Ok(true) = plugins.contains_key(plugin.id.clone()) {
					warn!("a plugin with id {} is already registered, skipping", plugin.id)
				} else {
					info!("loaded plugin {}", plugin.id);
					plugins.set(plugin.id.clone(), plugin.table)?;
					priorities.push((plugin.priority, plugin.id));
				}
			}
			Err(e) => warn!("error loading plugin at {spath}: {e}"),
		}
	}
	priorities.sort_by_cached_key(|(i, _)| -i);
	debug!("plugin priorities: {priorities:?}");
	let priorities = priorities.into_iter()
		.map(|(_, id)| id).collect();
	Ok((plugins, priorities))
}
