use std::{collections::{HashMap, hash_map::Entry}, sync::Arc};

use log::{Record, debug};
use mlua::{Value, IntoLua, Lua, Table, Function, AnyUserData, LuaSerdeExt};
use serde_json::json;
use tokio::sync::{mpsc::Sender, RwLock};
use uuid::Uuid;

use crate::{QC_VERSION, JsonValue, command::Commands};

use super::{event::{PluginEvent, ConnectionInfo}, UuidUD};

impl<'lua> IntoLua<'lua> for ConnectionInfo {
    fn into_lua(self, lua: &'lua mlua::Lua) -> mlua::Result<Value<'lua>> {
        let table = lua.create_table()?;
		table.set("server_addr", self.server_addr.0)?;
		table.set("server_port", self.server_addr.1)?;
		table.set("remote_addr", self.remote_addr.0)?;
		table.set("remote_port", self.remote_addr.1)?;
		table.set("proto_version", self.protocol.0)?;
		table.set("proto_name", self.protocol.1)?;
		Ok(Value::Table(table))
    }
}

fn lua_locinfo(lua: &Lua) -> Option<(u32, String)> {
	let debug = lua.globals().get::<_, Table>("debug").ok()?;
	let getinfo = debug.get::<_, Function>("getinfo").ok()?;
	let data = getinfo.call::<_, Table>((2,)).ok()?;
	let line = data.get("currentline").ok()?;
	let file = data.get("short_src").ok()?;
	Some((line, file))
}

fn lua_log(level: log::Level, path: &str, msg: String, lua: &Lua) {
	let mut record = Record::builder();
	record.level(level);
	record.target(path);
	record.module_path(Some(path));
	let locinfo = lua_locinfo(lua);
	if let Some((line, file)) = &locinfo {
		record.line(Some(*line));
		record.file(Some(file));
	}
	log::logger().log(&record.args(format_args!("{}", msg)).build());
}

const LEVELS: [(&str, log::Level); 5] = [
	("trace", log::Level::Trace),
	("debug", log::Level::Debug),
	("info", log::Level::Info),
	("warn", log::Level::Warn),
	("error", log::Level::Error),
];

fn value_to_chat<'lua>(lua: &'lua Lua, value: Value<'lua>) -> mlua::Result<JsonValue> {
	match value {
		Value::Table(_) => Ok(lua.from_value(value)?),
		Value::String(s) => Ok(json!{{"text": s.to_str()?}}),
		Value::Nil => Ok(json!{{"text":""}}),
		_ => Err(mlua::Error::RuntimeError("()".to_owned())),
	}
}

fn opt_value_to_chat<'lua>(lua: &'lua Lua, value: Value<'lua>) -> mlua::Result<Option<JsonValue>> {
	match value {
		Value::Table(_) => Ok(Some(lua.from_value(value)?)),
		Value::String(s) => Ok(Some(json!{{"text": s.to_str()?}})),
		Value::Nil => Ok(None),
		_ => Err(mlua::Error::RuntimeError("()".to_owned())),
	}
}


pub struct Server {
	pub tx: Sender<PluginEvent>,
	pub commands: Arc<RwLock<Commands>>,
	pub command_ids: RwLock<HashMap<String, String>>,
	pub names: RwLock<HashMap<Uuid, String>>,
	pub info: RwLock<HashMap<Uuid, ConnectionInfo>>,
}

#[derive(Clone)]
pub struct WrappedServer(pub Arc<Server>);

impl std::ops::Deref for WrappedServer {
	type Target = Server;

	fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl mlua::UserData for WrappedServer {
	fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
		// commands
		methods.add_async_function("add_command", |_lua, (this, name): (AnyUserData, String)| async move {
			let id: String = this.nth_user_value(2)?;

			debug!("registering simple command '{name}' for plugin '{id}'");

			let this_srv: std::cell::Ref<Self> = this.borrow()?;

			let mut cmds_guard = this_srv.commands.write().await;
			let mut cmd_ids_guard = this_srv.command_ids.write().await;

			let qualified_name = id.clone() + ":" + &name;

			// unqualified name
			if let Some(other_id) = cmd_ids_guard.get(&name) {
				debug!("cannot register unqualified command '{name}' for plugin '{id}', as it is already in use by plugin '{other_id}'")
			} else {
				cmd_ids_guard.insert(name.clone(), id.clone());

				cmds_guard.add_simple_command(name)
					.map_err(mlua::Error::runtime)?;
			}

			// qualified name
			cmd_ids_guard.insert(qualified_name.clone(), id);

			cmds_guard.add_simple_command(qualified_name)
				.map_err(mlua::Error::runtime)?;

			drop(cmd_ids_guard);
			drop(cmds_guard);

			Ok(())
		});

		methods.add_async_function("register_command", |_lua, (this, name): (AnyUserData, String)| async move {
			let id: String = this.nth_user_value(2)?;
			let this_srv: std::cell::Ref<Self> = this.borrow()?;

			let mut cmd_ids_guard = this_srv.command_ids.write().await;

			let qualified_name = id.clone() + ":" + &name;
			cmd_ids_guard.insert(qualified_name, id.clone());

			if let Entry::Vacant(entry) = cmd_ids_guard.entry(name) {
				entry.insert(id.clone());
				Ok(true)
			} else {
				Ok(false)
			}
		});

		methods.add_async_method("update_commands", |_lua, this, ()| async {
			this.tx.send(PluginEvent::UpdateCommands).await.unwrap();
			Ok(())
		});

		// players
		methods.add_async_method("players", |lua, this, ()| async {
			let table = lua.create_table()?;
			for (k, v) in this.names.read().await.iter() {
				table.set(UuidUD(*k), v.clone())?;
			}
			Ok(table)
		});
		methods.add_async_method("get_name", |lua, this, UuidUD(uuid)| async move {
			match this.names.read().await.get(&uuid) {
				Some(s) => s.clone().into_lua(lua),
				None => Ok(Value::Nil),
			}
		});
		methods.add_async_method("get_uuid", |lua, this, name: String| async move {
			for (uuid, pname) in this.names.read().await.iter() {
				if pname.eq_ignore_ascii_case(&name) {
					return UuidUD(*uuid).into_lua(lua)
				}
			}
			Ok(Value::Nil)
		});
		methods.add_async_method("get_info", |lua, this, UuidUD(uuid)| async move {
			match this.info.read().await.get(&uuid) {
				Some(i) => i.clone().into_lua(lua),
				None => Ok(Value::Nil)
			}
		});

		// events
		methods.add_async_method("kick", |lua, this, (UuidUD(uuid), msg)| async move {
			let msg = opt_value_to_chat(lua, msg)?;
			this.tx
				.send(PluginEvent::Kick(uuid, msg))
				.await.unwrap();
			Ok(())
		});
		methods.add_async_method("send_msg", |lua, this, (msg, uuid): (Value, Option<UuidUD>)| async move {
			let msg = value_to_chat(lua, msg)?;
			this.tx
				.send(PluginEvent::System(msg, uuid.map(|u| u.0)))
				.await.unwrap();
			Ok(())
		});
		methods.add_async_method("send_chat", |lua, this, (msg, uuid): (Value, Option<UuidUD>)| async move {
			let msg = value_to_chat(lua, msg)?;
			this.tx
				.send(PluginEvent::Chat(msg, uuid.map(|u| u.0)))
				.await.unwrap();
			Ok(())
		});

		// logging
		for (name, level) in LEVELS {
			methods.add_function(name, move |lua, (this, msg): (AnyUserData, String)| {
				let mut id: String = this.nth_user_value(2)?;
				id.insert_str(0, "plugin::");
				lua_log(level, &id, msg, lua);
				Ok(())
			});
		}

		// utility
		methods.add_function("check_semver", |_, (ver, req): (String, String)| {
			let ver = semver::Version::parse(&ver)
				.map_err(|_| mlua::Error::runtime(format!("invalid version: {ver}")))?;
			let req = semver::VersionReq::parse(&req)
				.map_err(|_| mlua::Error::runtime(format!("invalid version request: {req}")))?;
			Ok(req.matches(&ver))
		});
		methods.add_function("parse_uuid", |lua, uuid: String| {
			match Uuid::parse_str(&uuid) {
				Ok(u) => UuidUD(u).into_lua(lua),
				Err(_) => Ok(Value::Nil),
			}
		});

	}

	fn add_fields<'lua, F: mlua::UserDataFields<'lua, Self>>(fields: &mut F) {
		fields.add_field("version", QC_VERSION);
		fields.add_field_function_get("plugins", |_, this| {
			this.nth_user_value::<Table>(1)
		})
	}
}

