use std::io::{Write, Read};

use log::trace;
use uuid::Uuid;

use crate::{ser::{Serializable, Deserializable}, varint::VarInt, command::Commands};

use super::{ServerPacket, ClientPacket, Protocol, ProtocolState, common::COMMON};

pub const NAME: &str = "1.20.1";
pub const VERSION: i32 = 763;

pub const PROTOCOL: Protocol = Protocol {
    name: NAME,
    version: VERSION,
    encode,
    decode,
};

fn encode(mut w: Box<&mut dyn Write>, state: ProtocolState, ev: ServerPacket) -> anyhow::Result<()> {
    trace!("{NAME} encoding {ev:?}");
    match ev {
        ServerPacket::LoginSuccess { name, uuid } => {
            VarInt(0x02).serialize(&mut w)?;
            uuid.serialize(&mut w)?;
            name.serialize(&mut w)?;
            // number of properties (0)
            VarInt(0x00).serialize(&mut w)?;
        },
        ServerPacket::CommandData(commands) => {
            VarInt(0x10).serialize(&mut w)?;
            VarInt(commands.len()).serialize(&mut w)?;
            commands.serialize(&mut w, true)?;
            VarInt(Commands::root()).serialize(&mut w)?;
        },
        ServerPacket::PluginMessage { channel, data } => {
            VarInt(0x17).serialize(&mut w)?;
            channel.serialize(&mut w)?;
            w.write_all(&data)?;
        },
        ServerPacket::PlayDisconnect(msg) => {
            VarInt(0x1A).serialize(&mut w)?;
            msg.serialize(&mut w)?;
        },
        ServerPacket::ChatMessage(msg, name) => {
            VarInt(0x1B).serialize(&mut w)?;
            msg.serialize(&mut w)?;
            VarInt(7).serialize(&mut w)?; // chat type from registry codec
            name.serialize(&mut w)?;
            false.serialize(&mut w)?; // no target
        },
        ServerPacket::KeepAlive(data) => {
            VarInt(0x23).serialize(&mut w)?;
            data.serialize(&mut w)?;
        },
        ServerPacket::ChunkData { x, z } => {
            VarInt(0x24).serialize(&mut w)?;
            // chunk x
            x.serialize(&mut w)?;
            // chunk z
            z.serialize(&mut w)?;
            // heightmap
            include_bytes!("../resources/heightmap.nbt").serialize(&mut w)?;
            // chunk data size and chunk data
            VarInt(192).serialize(&mut w)?;
            [0u8; 192].serialize(&mut w)?;
            // block entities
            VarInt(0).serialize(&mut w)?;
            // light masks
            VarInt(0).serialize(&mut w)?;
            VarInt(0).serialize(&mut w)?;
            VarInt(0).serialize(&mut w)?;
            VarInt(0).serialize(&mut w)?;
            // sky light array len
            VarInt(0).serialize(&mut w)?;
            // block light array len
            VarInt(0).serialize(&mut w)?;
        },
        ServerPacket::JoinGame { eid, gamemode, hardcode } => {
            VarInt(0x28).serialize(&mut w)?;
            eid.serialize(&mut w)?;
            hardcode.serialize(&mut w)?;
            gamemode.serialize(&mut w)?;
            (-1i8).serialize(&mut w)?; // prev gamemode undefined
            VarInt(1).serialize(&mut w)?; // dimension count
            "qc:world".serialize(&mut w)?; // register one dimension
            include_bytes!("../resources/registry_1.20.1.nbt").serialize(&mut w)?; // registry codec
            "minecraft:the_end".serialize(&mut w)?; // dimension type
            "qc:world".serialize(&mut w)?; // dimension name
            0i64.serialize(&mut w)?; // seed
            VarInt(65535).serialize(&mut w)?; // max players
            VarInt(8).serialize(&mut w)?; // view dist
            VarInt(0).serialize(&mut w)?; // sim dist
            false.serialize(&mut w)?; // reduce debug info
            false.serialize(&mut w)?; // respawn screen
            false.serialize(&mut w)?; // is debug
            false.serialize(&mut w)?; // is flat
            false.serialize(&mut w)?; // has death location
            VarInt(1).serialize(&mut w)?; // portal cooldown
        },
        ServerPacket::PositionAndLook { pos, look, flags } => {
            VarInt(0x3C).serialize(&mut w)?;
            pos.serialize(&mut w)?;
            look.serialize(&mut w)?;
            flags.serialize(&mut w)?;
            VarInt(0).serialize(&mut w)?;
        },
        ServerPacket::SetDefaultSpawn { pos, angle } => {
            VarInt(0x50).serialize(&mut w)?;
            pos.serialize(&mut w)?;
            angle.serialize(&mut w)?;
        },
        ServerPacket::SystemMessage(msg, overlay) => {
            VarInt(0x64).serialize(&mut w)?;
            msg.serialize(&mut w)?;
            overlay.serialize(&mut w)?;
        },
        _ => { (COMMON.encode)(w, state, ev)?; }
    }
    Ok(())
}

fn decode(mut r: Box<&mut dyn Read>, state: ProtocolState, len: i32, id: i32) -> anyhow::Result<Option<ClientPacket>> {
    trace!("{NAME} decoding {state:?} {id}");
    type Ps = ProtocolState;
    match (state, id) {
        (Ps::Login, 0x00) => {
            let name = String::deserialize(&mut r)?;
            let has_uuid = bool::deserialize(&mut r)?;
            let uuid = if has_uuid {
                Some(Uuid::deserialize(&mut r)?)
            } else { None };
            Ok(Some(ClientPacket::LoginStart { name, uuid }))
        }
        (Ps::Login, 0x01 | 0x02) => Ok(None), // unsupported
        (Ps::Play, 0x04) => {
            let cmd = String::deserialize(&mut r)?;
            Ok(Some(ClientPacket::Command(cmd)))
        },
        (Ps::Play, 0x05) => {
            let msg = String::deserialize(&mut r)?;
            Ok(Some(ClientPacket::ChatMessage(msg)))
        },
        (Ps::Play, 0x0D) => {
            let channel = String::deserialize(&mut r)?;
            let mut data = Vec::new();
            r.read_to_end(&mut data)?;
            Ok(Some(ClientPacket::PluginMessage { channel, data }))
        },
        (Ps::Play, 0x12) => {
            let data = i64::deserialize(&mut r)?;
            Ok(Some(ClientPacket::KeepAlive(data)))
        },
        (Ps::Play, 0x14 | 0x15 | 0x16) => Ok(None), // position & rotation
        _ => (COMMON.decode)(r, state, len, id)
    }
}
