use std::{io::{self, Read, Write}, mem::MaybeUninit};

use anyhow::anyhow;
use uuid::Uuid;

use crate::JsonValue;

use super::varint::{VarInt, VarLong};

pub trait ReadExt: Read {
	fn readb(&mut self) -> io::Result<u8> {
		Ok(self.readn::<1>()?[0])
	}

	fn readn<const N: usize>(&mut self) -> io::Result<[u8; N]> {
		let mut buf = [0; N];
		self.read_exact(&mut buf)?;
		Ok(buf)
	}
}

impl<T: Read> ReadExt for T {}

pub trait Serializable: Sized {
	fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()>;
}

pub trait Deserializable: Sized {
	fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self>;
}

impl Serializable for bool {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
		w.write_all(&[*self as u8])?;
		Ok(())
    }
}

impl Deserializable for bool {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
		match r.readb()? {
			0 => Ok(false),
			1 => Ok(true),
			b => Err(anyhow!("value {b} out of range for bool"))
		}
    }
}

macro_rules! serialize_primitive {
	($ty:ty) => {
		impl Serializable for $ty {
			fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
				w.write_all(&self.to_be_bytes())?;
				Ok(())
			}
		}

		impl Deserializable for $ty {
			fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
				Ok(Self::from_be_bytes(r.readn()?))
			}
		}
	};
}

serialize_primitive!(i8);
serialize_primitive!(i16);
serialize_primitive!(i32);
serialize_primitive!(i64);
serialize_primitive!(i128);
serialize_primitive!(u8);
serialize_primitive!(u16);
serialize_primitive!(u32);
serialize_primitive!(u64);
serialize_primitive!(u128);
serialize_primitive!(f32);
serialize_primitive!(f64);

impl Serializable for VarInt {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
        let mut data = self.0 as u32;
        loop {
            let mut byte = (data & 0x7F) as u8;
            data >>= 7;
            if data != 0 {
                byte |= 0x80;
            }
            w.write_all(&[byte])?;
            if data == 0 {
                return Ok(())
            }
        }
    }
}

impl Deserializable for VarInt {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
		let mut result = 0u32;
        for count in 0..5 {
            let byte = r.readb()?;
            result |= ((byte & 0x7f) as u32) << (7 * count);
            if byte & 0x80 == 0 {
                return Ok(VarInt(result as i32))
            }
        }
		Err(anyhow!("VarInt too long"))
    }
}

impl Serializable for VarLong {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
        let mut data = self.0 as u64;
        loop {
            let mut byte = (data & 0x7F) as u8;
            data >>= 7;
            if data != 0 {
                byte |= 0x80;
            }
            w.write_all(&[byte])?;
            if data == 0 {
                return Ok(())
            }
        }
    }
}

impl Deserializable for VarLong {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
		let mut result = 0u64;
		for count in 0..10 {
            let byte = r.readb()?;
            result |= ((byte & 0x7f) as u64) << (7 * count);
            if byte & 0x80 == 0 {
                return Ok(VarLong(result as i64))
            }
        }
		Err(anyhow!("VarLong too long"))
    }
}

impl Serializable for String {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
		VarInt(self.len() as i32).serialize(w)?;
		w.write_all(self.as_bytes())?;
		Ok(())
    }
}

impl Deserializable for String {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
		let len = VarInt::deserialize(r)?.0 as u64;
		let mut buf = String::new();
		r.take(len).read_to_string(&mut buf)?;
		Ok(buf)
    }
}

impl Serializable for &str {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
		VarInt(self.len() as i32).serialize(w)?;
		w.write_all(self.as_bytes())?;
		Ok(())
    }
}

impl Serializable for JsonValue {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
		serde_json::to_string(self)?.serialize(w)
    }
}

impl Deserializable for JsonValue {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
		let v = serde_json::from_str(&String::deserialize(r)?)?;
		Ok(v)
    }
}

impl<T, const N: usize> Serializable for [T; N]
where T: Serializable {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
		for item in self {
			item.serialize(w)?;
		}
		Ok(())
    }
}
impl<T, const N: usize> Deserializable for [T; N]
where T: Deserializable {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
		let mut arr: [MaybeUninit<T>; N] = unsafe {
			MaybeUninit::uninit().assume_init()
		};
		for item in arr.iter_mut().take(N) {
			*item = MaybeUninit::new(T::deserialize(r)?);
		}
		// Safety: since MaybeUninit<T> and T have the same memory
		// layout, so will [MaybeUninit<T>; N] and [T; N].
		// std::mem::transmute cannot yet handle arrays of
		// generic size.
		Ok(unsafe { arr.as_ptr().cast::<Self>().read() })
    }
}

impl Serializable for Uuid {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
        self.as_u128().serialize(w)
    }
}

impl Deserializable for Uuid {
    fn deserialize(r: &mut impl ReadExt) -> anyhow::Result<Self> {
        Ok(Uuid::from_u128(u128::deserialize(r)?))
    }
}

#[derive(Clone, Copy, Debug)]
pub struct Position {
	pub x: i32,
	pub y: i16,
	pub z: i32,
}


impl Serializable for Position {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
        let value = (((self.x & 0x3FFFFFF) as u64) << 38)
            | (((self.z & 0x3FFFFFF) as u64) << 12)
            | ((self.y & 0xFFF) as u64);
		value.serialize(w)
    }
}

#[derive(Clone, Copy, Debug)]
pub struct PositionOld {
    pub x: i32,
    pub z: i32,
    pub y: i16,
}

impl Serializable for PositionOld {
    fn serialize(&self, w: &mut impl Write) -> anyhow::Result<()> {
        let value = (((self.x & 0x3FFFFFF) as u64) << 38)
            | (((self.y & 0xFFF) as u64) << 26)
            | ((self.z & 0x3FFFFFF) as u64);
        value.serialize(w)
    }
}

impl From<Position> for PositionOld {
    fn from(value: Position) -> Self {
        Self { x: value.x, y: value.y, z: value.z }
    }
}
