use std::{collections::HashMap, time::Duration, net::SocketAddr, sync::Arc};

use log::{info, error};
use serde_json::json;
use tokio::{net::{TcpListener, TcpStream}, sync::{mpsc::{Sender, self}, RwLock}, select};
use uuid::Uuid;
use anyhow::anyhow;

use crate::{client::{event::ClientInfo, run_client}, Player, client::event::{ServerEvent, ClientEvent}, plugin::{run_plugins, event::{PluginEvent, ConnectionInfo}}, command::Commands, CONFIG};

type PServerEvent = crate::plugin::event::ServerEvent;


struct Client {
	tx: Sender<ServerEvent>,
	remote_addr: SocketAddr,
	keepalive: Option<i64>,
	info: Option<ClientInfo>,
	player: Option<Player>,
}


struct Server {
	commands: Arc<RwLock<Commands>>,
	gc_tx: Sender<(i32, ClientEvent)>,
	ps_tx: Sender<PServerEvent>,
	next_id: i32,
	clients: HashMap<i32, Client>,
	ids: HashMap<Uuid, i32>,
}

pub async fn run_server() -> anyhow::Result<()> {
	env_logger::init();

	let (ps_tx, ps_rx) = mpsc::channel(256);
	let (pc_tx, mut pc_rx) = mpsc::channel(256);

	let commands = Arc::new(RwLock::new(Commands::new()));
	let pl_commands = commands.clone();

	tokio::spawn(async {
		if let Err(e) = run_plugins(pc_tx, ps_rx, pl_commands).await {
			error!("error in plugins: {e}");
		}
	});

	let server_addr = CONFIG.get().unwrap().address;
	let listener = TcpListener::bind(server_addr).await?;

	let (gc_tx, mut gc_rx) = mpsc::channel(16);

	let mut server = Server {
		commands,
		gc_tx, ps_tx,
		next_id: 0,
		clients: HashMap::new(),
		ids: HashMap::new(),
	};

	let mut keepalive = tokio::time::interval(Duration::from_secs(15));

	info!("listening on {server_addr}");

	loop { select! {
		conn = listener.accept() => {
			let (stream, addr) = conn?;
			accept_connection(&mut server, stream, addr).await;
		},
		_ = keepalive.tick() => handle_keepalive(&mut server).await?,
		ev = pc_rx.recv() => {
			let Some(ev) = ev else {
				return Err(anyhow!("plugin channel closed"));
			};
			handle_plugin_event(&mut server, ev).await?;
		},
		ev = gc_rx.recv() => {
			let Some(ev) = ev else {
				return Err(anyhow!("reciever closed"));
			};
			handle_client_event(&mut server, ev.0, ev.1).await?;
		}
	} }
}

async fn accept_connection(server: &mut Server, stream: TcpStream, addr: SocketAddr) {
	let gc_tx = server.gc_tx.clone();

	let id = loop {
		let id = server.next_id;
		server.next_id = server.next_id.wrapping_add(1);
		if !server.clients.contains_key(&id) {
			break id
		}
	};

	let (gs_tx, gs_rx) = mpsc::channel(16);
	server.clients.insert(id, Client {
		tx: gs_tx,
		remote_addr: addr,
		keepalive: None,
		info: None,
		player: None,
	});
	info!("#{id} connected from {addr}");
	tokio::spawn(async move {
		let c_tx2 = gc_tx.clone();
		match run_client(id, stream, gc_tx, gs_rx).await {
			Ok(_) => info!("#{id} disconnected"),
			Err(e) => error!("#{id} disconnected: {e}"),
		}
		c_tx2.send((id, ClientEvent::Disconnect)).await.unwrap();
	});
}

async fn handle_keepalive(server: &mut Server) -> anyhow::Result<()> {
	for client in server.clients.values_mut() {
		if client.keepalive.is_some() {
			client.tx.send(ServerEvent::Disconnect(None)).await?;
		} else if client.player.is_some() {
			let data = 1;
			client.keepalive = Some(data);
			client.tx.send(ServerEvent::KeepAlive(data)).await?;
		}
	}
	Ok(())
}

async fn handle_plugin_event(server: &mut Server, ev: PluginEvent) -> anyhow::Result<()> {
	match ev {
		PluginEvent::Kick(uuid, msg) => {
			if let Some(id) = server.ids.get(&uuid) {
				server.clients[id].tx
					.send(ServerEvent::Disconnect(msg))
					.await?;
			}
		},
		PluginEvent::Chat(msg, Some(uuid)) => {
			if let Some(id) = server.ids.get(&uuid) {
				server.clients[id].tx
					.send(ServerEvent::ChatMessage(msg, serde_json::json!({"text":""})))
					.await?;
			}
		},
		PluginEvent::Chat(msg, None) => {
			for client in server.clients.values() {
				if client.player.is_some() {
					client.tx.send(ServerEvent::ChatMessage(msg.clone(), serde_json::json!({"text":""}))).await?;
				}
			}
		},
		PluginEvent::System(msg, Some(uuid)) => {
			if let Some(id) = server.ids.get(&uuid) {
				server.clients[id].tx
					.send(ServerEvent::SystemMessage(msg, false))
					.await?;
			}
		},
		PluginEvent::System(msg, None) => {
			for client in server.clients.values() {
				if client.player.is_some() {
					client.tx.send(ServerEvent::SystemMessage(msg.clone(), false)).await?;
				}
			}
		},
		PluginEvent::UpdateCommands => {
			let commands = server.commands.read().await.clone();
			for client in server.clients.values() {
				if client.player.is_some() {
					client.tx.send(ServerEvent::UpdateCommands(commands.clone())).await?;
				}
			}
		}
	}
	Ok(())
}

async fn handle_client_event(server: &mut Server, id: i32, ev: ClientEvent) -> anyhow::Result<()> {
	match ev {
		ClientEvent::Handshake(info) => {
			server.clients.get_mut(&id).unwrap().info = Some(info)
		},
		ClientEvent::Join(player) => {
			if let Some(other_id) = server.ids.remove(&player.uuid) {
				let other_client = server.clients.remove(&other_id).unwrap();
				server.ps_tx
					.send(PServerEvent::Leave(other_client.player.unwrap()))
					.await?;
				other_client.tx
					.send(ServerEvent::Disconnect(Some(json!({"translate": "multiplayer.disconnect.duplicate_login"}))))
					.await?;
			}
			let client = server.clients.get_mut(&id).unwrap();
			client.player = Some(player.clone());
			client.tx.send(ServerEvent::UpdateCommands(server.commands.read().await.clone())).await?;
			server.ids.insert(player.uuid, id);
			let info = client.info.as_ref().unwrap();
			let conn_info = ConnectionInfo {
				server_addr: (info.addr.to_owned(), info.port),
				remote_addr: (client.remote_addr.ip().to_string(), client.remote_addr.port()),
				protocol: (info.proto_version, info.proto_name.to_owned())
			};
			server.ps_tx.send(PServerEvent::Join(player, conn_info)).await?;
		},
		ClientEvent::Disconnect => if let Some(client) = server.clients.remove(&id) {
			if let Some(player) = client.player {
				server.ids.remove(&player.uuid);
				server.ps_tx.send(PServerEvent::Leave(player)).await?;
			}
		},
		ClientEvent::KeepAlive(data) => {
			let client = server.clients.get_mut(&id).unwrap();
			if client.keepalive == Some(data) {
				client.keepalive = None;
			} else {
				client.keepalive = Some(0);
			}
		},
		ClientEvent::Command(cmd) => {
			let player = server.clients.get(&id).unwrap().player.clone().unwrap();
			server.ps_tx.send(PServerEvent::Command(player, cmd)).await?;
		},
		ClientEvent::Message(msg) => {
			let player = server.clients.get(&id).unwrap().player.clone().unwrap();
			server.ps_tx.send(PServerEvent::Message(player, msg)).await?;
		},
	}
	Ok(())
}
